@@ -4,6 +4,8 @@ from numba import jit
 # [1] Arcolezi et al (2021) "Improving the Utility of Locally Differentially Private Protocols for Longitudinal and Multidimensional Frequency Estimates" (arXiv:2111.04636).
 # [2] Wang et al (2017) "Locally differentially private protocols for frequency estimation" (USENIX Security).
 
+# The analytical analysis of how to calculate parameters (p1, q2, p2, q2) is from: https://github.com/hharcolezi/ldp-protocols-mobility-cdrs/blob/main/papers/%5B4%5D/1_ALLOMFREE_Analysis.ipynb
+
 @jit(nopython=True)
 def L_SOUE_Client(input_data, k, eps_perm, eps_1):
 
@@ -16,74 +18,78 @@ def L_SOUE_Client(input_data, k, eps_perm, eps_1):
     :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
     :return: sanitized value or UE vector.
     """
-
-    # SUE parameters for round 1
-    p1 = np.exp(eps_perm / 2) / (np.exp(eps_perm / 2) + 1)
-    q1 = 1 - p1
-
-    # OUE parameters for round 2
-    p2 = 0.5
-    q2 = (3.35410196624968 * (np.exp(eps_1) - 1) * (
-                np.exp(0.5 * eps_perm) + 2 * np.exp(eps_perm) + np.exp(1.5 * eps_perm)) * np.sqrt(
-        0.0111111111111111 * np.exp(eps_1) + 0.00555555555555556 * np.exp(2 * eps_1) + 0.0444444444444444 * np.exp(
-            0.5 * eps_perm) + 0.155555555555556 * np.exp(eps_perm) + 0.311111111111111 * np.exp(
-            1.5 * eps_perm) + 0.388888888888889 * np.exp(2 * eps_perm) + 0.311111111111111 * np.exp(
-            2.5 * eps_perm) + 0.155555555555556 * np.exp(3 * eps_perm) + 0.0444444444444444 * np.exp(
-            3.5 * eps_perm) + 0.00555555555555556 * np.exp(4 * eps_perm) - 0.222222222222222 * np.exp(
-            eps_1 + eps_perm) - 0.711111111111111 * np.exp(eps_1 + 1.5 * eps_perm) - np.exp(
-            eps_1 + 2 * eps_perm) - 0.711111111111111 * np.exp(eps_1 + 2.5 * eps_perm) - 0.222222222222222 * np.exp(
-            eps_1 + 3 * eps_perm) + 0.0111111111111111 * np.exp(eps_1 + 4 * eps_perm) + 0.0444444444444444 * np.exp(
-            2 * eps_1 + 0.5 * eps_perm) + 0.155555555555556 * np.exp(2 * eps_1 + eps_perm) + 0.311111111111111 * np.exp(
-            2 * eps_1 + 1.5 * eps_perm) + 0.388888888888889 * np.exp(
-            2 * eps_1 + 2 * eps_perm) + 0.311111111111111 * np.exp(
-            2 * eps_1 + 2.5 * eps_perm) + 0.155555555555556 * np.exp(
-            2 * eps_1 + 3 * eps_perm) + 0.0444444444444444 * np.exp(
-            2 * eps_1 + 3.5 * eps_perm) + 0.00555555555555556 * np.exp(
-            2 * eps_1 + 4 * eps_perm) + 0.00555555555555556) - 0.25 * (
-                      np.exp(0.5 * eps_perm) + 2 * np.exp(eps_perm) + np.exp(1.5 * eps_perm) - np.exp(
-                  eps_1 + 0.5 * eps_perm) - 2 * np.exp(eps_1 + eps_perm) - np.exp(eps_1 + 1.5 * eps_perm)) * (
-                      np.exp(eps_1) + 4 * np.exp(0.5 * eps_perm) + 4 * np.exp(eps_perm) - np.exp(
-                  2 * eps_perm) - 4 * np.exp(eps_1 + eps_perm) - 4 * np.exp(eps_1 + 1.5 * eps_perm) - np.exp(
-                  eps_1 + 2 * eps_perm) + 1)) / (
-                     (np.exp(eps_1) - 1) * (np.exp(0.5 * eps_perm) + 2 * np.exp(eps_perm) + np.exp(1.5 * eps_perm)) * (
-                         np.exp(0.5 * eps_perm) + 2 * np.exp(eps_perm) + np.exp(1.5 * eps_perm) - np.exp(
-                     eps_1 + 0.5 * eps_perm) - 2 * np.exp(eps_1 + eps_perm) - np.exp(eps_1 + 1.5 * eps_perm)))
-
-    if (np.array([p1, q1, p2, q2]) >= 0).all():
-        pass
+    
+    if eps_1 >= eps_perm:
+        raise ValueError('Please set eps_1 (single report, i.e., lower bound) < eps_perm (infinity reports, i.e., upper bound)')
+    
     else:
-        raise ValueError('Probabilities are negative.')
+        # SUE parameters for round 1
+        p1 = np.exp(eps_perm / 2) / (np.exp(eps_perm / 2) + 1)
+        q1 = 1 - p1
 
-    # Unary encoding
-    input_ue_data = np.zeros(k)
-    if input_data != None:
-        input_ue_data[input_data] = 1
+        # OUE parameters for round 2
+        p2 = 0.5
+        q2 = (3.35410196624968 * (np.exp(eps_1) - 1) * (
+                    np.exp(0.5 * eps_perm) + 2 * np.exp(eps_perm) + np.exp(1.5 * eps_perm)) * np.sqrt(
+            0.0111111111111111 * np.exp(eps_1) + 0.00555555555555556 * np.exp(2 * eps_1) + 0.0444444444444444 * np.exp(
+                0.5 * eps_perm) + 0.155555555555556 * np.exp(eps_perm) + 0.311111111111111 * np.exp(
+                1.5 * eps_perm) + 0.388888888888889 * np.exp(2 * eps_perm) + 0.311111111111111 * np.exp(
+                2.5 * eps_perm) + 0.155555555555556 * np.exp(3 * eps_perm) + 0.0444444444444444 * np.exp(
+                3.5 * eps_perm) + 0.00555555555555556 * np.exp(4 * eps_perm) - 0.222222222222222 * np.exp(
+                eps_1 + eps_perm) - 0.711111111111111 * np.exp(eps_1 + 1.5 * eps_perm) - np.exp(
+                eps_1 + 2 * eps_perm) - 0.711111111111111 * np.exp(eps_1 + 2.5 * eps_perm) - 0.222222222222222 * np.exp(
+                eps_1 + 3 * eps_perm) + 0.0111111111111111 * np.exp(eps_1 + 4 * eps_perm) + 0.0444444444444444 * np.exp(
+                2 * eps_1 + 0.5 * eps_perm) + 0.155555555555556 * np.exp(2 * eps_1 + eps_perm) + 0.311111111111111 * np.exp(
+                2 * eps_1 + 1.5 * eps_perm) + 0.388888888888889 * np.exp(
+                2 * eps_1 + 2 * eps_perm) + 0.311111111111111 * np.exp(
+                2 * eps_1 + 2.5 * eps_perm) + 0.155555555555556 * np.exp(
+                2 * eps_1 + 3 * eps_perm) + 0.0444444444444444 * np.exp(
+                2 * eps_1 + 3.5 * eps_perm) + 0.00555555555555556 * np.exp(
+                2 * eps_1 + 4 * eps_perm) + 0.00555555555555556) - 0.25 * (
+                          np.exp(0.5 * eps_perm) + 2 * np.exp(eps_perm) + np.exp(1.5 * eps_perm) - np.exp(
+                      eps_1 + 0.5 * eps_perm) - 2 * np.exp(eps_1 + eps_perm) - np.exp(eps_1 + 1.5 * eps_perm)) * (
+                          np.exp(eps_1) + 4 * np.exp(0.5 * eps_perm) + 4 * np.exp(eps_perm) - np.exp(
+                      2 * eps_perm) - 4 * np.exp(eps_1 + eps_perm) - 4 * np.exp(eps_1 + 1.5 * eps_perm) - np.exp(
+                      eps_1 + 2 * eps_perm) + 1)) / (
+                         (np.exp(eps_1) - 1) * (np.exp(0.5 * eps_perm) + 2 * np.exp(eps_perm) + np.exp(1.5 * eps_perm)) * (
+                             np.exp(0.5 * eps_perm) + 2 * np.exp(eps_perm) + np.exp(1.5 * eps_perm) - np.exp(
+                         eps_1 + 0.5 * eps_perm) - 2 * np.exp(eps_1 + eps_perm) - np.exp(eps_1 + 1.5 * eps_perm)))
 
-    # First round of sanitization (permanent memoization) with SUE using user's input_ue_data
-    first_sanitization = np.zeros(k)
-    for ind in range(k):
-        if input_ue_data[ind] != 1:
-            rnd = np.random.random()
-            if rnd <= q1:
-                first_sanitization[ind] = 1
+        if (np.array([p1, q1, p2, q2]) >= 0).all():
+            pass
         else:
-            rnd = np.random.random()
-            if rnd <= p1:
-                first_sanitization[ind] = 1
+            raise ValueError('Probabilities are negative, selecting eps_1 << eps_perm might probably solve it.')
 
-    # Second round of sanitization with OUE using first_sanitization as input
-    second_sanitization = np.zeros(k)
-    for ind in range(k):
-        if first_sanitization[ind] != 1:
-            rnd = np.random.random()
-            if rnd <= q2:
-                second_sanitization[ind] = 1
-        else:
-            rnd = np.random.random()
-            if rnd <= p2:
-                second_sanitization[ind] = 1
+        # Unary encoding
+        input_ue_data = np.zeros(k)
+        if input_data != None:
+            input_ue_data[input_data] = 1
 
-    return second_sanitization
+        # First round of sanitization (permanent memoization) with SUE using user's input_ue_data
+        first_sanitization = np.zeros(k)
+        for ind in range(k):
+            if input_ue_data[ind] != 1:
+                rnd = np.random.random()
+                if rnd <= q1:
+                    first_sanitization[ind] = 1
+            else:
+                rnd = np.random.random()
+                if rnd <= p1:
+                    first_sanitization[ind] = 1
+
+        # Second round of sanitization with OUE using first_sanitization as input
+        second_sanitization = np.zeros(k)
+        for ind in range(k):
+            if first_sanitization[ind] != 1:
+                rnd = np.random.random()
+                if rnd <= q2:
+                    second_sanitization[ind] = 1
+            else:
+                rnd = np.random.random()
+                if rnd <= p2:
+                    second_sanitization[ind] = 1
+
+        return second_sanitization
 
 def L_SOUE_Aggregator(ue_reports, eps_perm, eps_1):
     """
